#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <regex.h>

#include "config.h"
#include "imapfilter.h"


extern unsigned int options;
extern unsigned int capabilities;

char *vbuf = NULL;
size_t vbufs = 0;

/*
 * Read one packet of data that the server sent.
 */
void receive_response(char *buf)
{
    socket_read(buf);

#ifdef DEBUG
    printf("\n%s\n", buf);
#endif

}


/*
 * Get server response to client's request.
 */
int server_response(unsigned int tag)
{
    char buf[RESPONSE_BUF];
    
    reset_vbuf();

    do {
	receive_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (tag && !strcasestr(buf, ultostr(tag, 16)));

    return analyze_response(buf);
}


/*
 * Process the greeting that server sends during connection.
 */
int greeting_response(void)
{
    char buf[RESPONSE_BUF];

    receive_response(buf);

    if (strcasestr(buf, "BYE"))
	return RESPONSE_BYE;
    else if (strcasestr(buf, "PREAUTH"))
	return RESPONSE_PREAUTH;

    return RESPONSE_OK;
}


/*
 * Process the data that server sent due to IMAP CAPABILITY client request.
 */
int capability_response(unsigned int tag)
{
    char buf[RESPONSE_BUF];

    reset_vbuf();

    do {
	receive_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));
    
    if (!strcasestr(vbuf, "IMAP4rev1")) {
	error("imapfilter: server does not support IMAP4rev1 protocol\n");
	return -2;
    }
    if (strcasestr(vbuf, "NAMESPACE"))
	capabilities |= CAPABILITY_NAMESPACE;

    return analyze_response(vbuf);
}


/*
 * Process the data that server sent due to IMAP NAMESPACE client request.
 */
int namespace_response(unsigned int tag, struct namespace_t * namesp)
{
    char buf[RESPONSE_BUF];
    char *c, *d;
    
    reset_vbuf();

    do {
	receive_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((c = strcasestr(vbuf, "* NAMESPACE"))) {
	c += 12;
	if (strncasecmp(c, "NIL", 3)) {
	    c = strchr(c, '"') + 1;
	    d = strchr(c, '"') + 1;

	    strncat(namesp->prefix, c, d - c - 1);
	    namesp->delim = *(strchr(d, '"') + 1);
	}
    }
    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP STATUS client request.
 */
int status_response(unsigned int tag, char *mbox)
{
    int r;
    char buf[RESPONSE_BUF];
    unsigned int exist, recent, unseen;
    char *c;

    exist = recent = unseen = 0;

    reset_vbuf();
    
    do {
	receive_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    r = analyze_response(buf);

    if (r == RESPONSE_NO)
	return -2;

    if ((c = strcasestr(vbuf, "MESSAGES"))) {
	c += 9;
	exist = strtoul(c, NULL, 10);
    }
    if ((c = strcasestr(vbuf, "RECENT"))) {
	c += 7;
	recent = strtoul(c, NULL, 10);
    }
    if ((c = strcasestr(vbuf, "UNSEEN"))) {
	c += 7;
	unseen = strtoul(c, NULL, 10);
    }
    if (!exist) {
	info("No messages in mailbox \"%s\".\n", mbox);
	return -2;
    }
    info("%d message%s, %d recent, %d unseen, in mailbox \"%s\".\n", exist,
	 plural(exist), recent, unseen, mbox);

    return r;
}


/*
 * Process the data that server sent due to IMAP SELECT client request.
 */
int select_response(unsigned int tag)
{
    char buf[RESPONSE_BUF];

    reset_vbuf();
    
    do {
	receive_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if (strcasestr(vbuf, "[READ-ONLY]"))
	return RESPONSE_READONLY;

    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP SEARCH client request.
 */
int search_response(unsigned int tag, char **mesgs)
{
   char buf[RESPONSE_BUF];
    char *c, *m;
    unsigned int blen;

    reset_vbuf();
    
    do {
	receive_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((c = strcasestr(vbuf, "* SEARCH "))) {
	blen = strlen(vbuf);
	
	m = *mesgs = (char *) xmalloc(blen + 1);
	
	c += 9;
	
	while (*c && (isdigit(*c) || *c == ' '))
	    *(m++) = *(c++);

	*m = 0;
    }

    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP FETCH client request.
 */
int fetch_response(unsigned int tag)
{
    char buf[RESPONSE_BUF];
    char hdrs[HEADERS_BUF];
    int f;
    unsigned int sa, sb, i, n;
    char s[8];
    char *b;

    hdrs[0] = 0;
    sa = sb = i = f = 0;
    s[0] = 0;

    do {
	receive_response(buf);

	b = buf;

	while (f || (b = strchr(b, '{'))) {
	    /* Headers are continued from previous packet. */
	    if (f) {
		f = 0;
		for (; i < HEADERS_BUF - 1 && i < sa - 2; i++) {
		    /* Continued in next packet. */
		    if (!*b) {
			f = 1;
			break;
		    }
		    hdrs[i] = *(b++);
		}
	    } else {
		/* Left bracket is last byte of buffer. */
		if (b == buf + RESPONSE_BUF - 2) {
		    receive_response(buf);
		    b = buf;
		} else
		    b++;

		sa = atoi(b);

		/* Headers length may be continued in next packet. */
		if (!(b = strchr(b, '}'))) {
		    receive_response(buf);
		    if (buf != (b = strchr(buf, '}'))) {
			sb = atoi(buf);
			strncat(s, ultostr(sa, 10), 7);
			strncat(s, ultostr(sb, 10), 7);
			sa = atoi(s);
			s[0] = 0;
		    }
		    b += 3;
		} else if ((n = RESPONSE_BUF - 1 - (b - buf)) < 3) {
		    receive_response(buf);
		    b = buf + 3 - n;
		} else {
		    b += 3;
		}

		for (i = 0; i < HEADERS_BUF - 1 && i < sa - 2; i++) {
		    if (!*b) {
			f = 1;
			break;
		    }
		    hdrs[i] = *(b++);
		}
	    }

	    if (!f) {
		hdrs[i] = 0;

		if (*hdrs) {
		    if (options & OPTION_HEADERS)
			info("%s\n", hdrs);
		    log_info(LOG_WRITE, hdrs);
		} else {
		    log_info(LOG_WRITE, NULL);
		}
	    } else {
		break;
	    }
	}
    } while (!strcasestr(buf, ultostr(tag, 16)));

    return analyze_response(buf);
}


/*
 * Process the data that server sent due to IMAP COPY client request.
 */
int copy_response(unsigned int tag)
{
    int r = RESPONSE_OK;
    char buf[RESPONSE_BUF];

    reset_vbuf();
    
    do {
	receive_response(buf);
	check_vbuf(strlen(buf));
	strncat(vbuf, buf, vbufs - strlen(vbuf));
    } while (!strcasestr(buf, ultostr(tag, 16)));

    if ((r = analyze_response(buf)) == RESPONSE_NO
	&& strcasestr(vbuf, "[TRYCREATE]"))
	return RESPONSE_TRYCREATE;
    
    return r;
}


/*
 * Check if response of server to client's request was succesfully
 * delivered or there was some kind of error.
 */
int analyze_response(char *buf)
{
    int r = RESPONSE_OK;
    regex_t creg;
    regmatch_t match[3];
    const char *reg = "[[:xdigit:]]{6,6} ((OK|NO|BAD)[[:print:]]*)\r\n";
    char result[RESULT_BUF];

    result[0] = 0;

    regcomp(&creg, reg, REG_EXTENDED | REG_ICASE);

    if (!regexec(&creg, buf, 3, match, 0)) {
	strncat(result, buf + match[1].rm_so,
		min(match[1].rm_eo - match[1].rm_so, RESULT_BUF - 1));

	if (!strncasecmp(buf + match[2].rm_so, "NO", 2))
	    r = RESPONSE_NO;
	else if (!strncasecmp(buf + match[2].rm_so, "BAD", 3))
	    r = RESPONSE_BAD;

	verbose("Server response: %s\n", result);
    }
    regfree(&creg);

    return r;
}


/*
 * Initialize virtual buffer. 
 */
void init_vbuf(void)
{
    vbuf = xmalloc(4096);
    *vbuf = 0;
    vbufs = 4096;
}


/*
 * Reset virtual buffer.
 */
void reset_vbuf(void)
{
    *vbuf = 0;
}

    
/*
 * Check if virtual buffer is full and make it bigger.
 */
void check_vbuf(size_t s)
{
    if (s + strlen(vbuf) >= vbufs) {
	vbufs += 4096;
	vbuf = xrealloc(vbuf, vbufs);
    }
}
       
